@@ -3,11 +3,11 @@
 from sys import version_info
 from unittest import TestCase
 
-import sys
+import pytest
 
-from parsimonious.exceptions import UndefinedLabel, ParseError
+from parsimonious.exceptions import BadGrammar, LeftRecursionError, ParseError, UndefinedLabel, VisitationError
 from parsimonious.expressions import Literal, Lookahead, Regex, Sequence, TokenMatcher, is_callable
-from parsimonious.grammar import rule_grammar, RuleVisitor, Grammar, TokenGrammar, LazyReference
+from parsimonious.grammar import rule_grammar, rule_syntax, RuleVisitor, Grammar, TokenGrammar, LazyReference
 from parsimonious.nodes import Node
 from parsimonious.utils import Token
 
@@ -285,6 +285,31 @@ class GrammarTests(TestCase):
             """)
         grammar.parse('(34)')
 
+    def test_resolve_refs_completeness(self):
+        """Smoke-test another circumstance where lazy references don't get resolved."""
+        grammar = Grammar(r"""
+            block = "{" _ item* "}" _
+
+            # An item is an element of a block.
+            item = number / word / block / paren
+
+            # Parens are for delimiting subexpressions.
+            paren = "(" _ item* ")" _
+
+            # Words are barewords, unquoted things, other than literals, that can live
+            # in lists. We may renege on some of these chars later, especially ".". We
+            # may add Unicode.
+            word = spaceless_word _
+            spaceless_word = ~r"[-a-z`~!@#$%^&*_+=|\\;<>,.?][-a-z0-9`~!@#$%^&*_+=|\\;<>,.?]*"i
+
+            number = ~r"[0-9]+" _ # There are decimals and strings and other stuff back on the "parsing" branch, once you get this working.
+
+            _ = meaninglessness*
+            meaninglessness = whitespace
+            whitespace = ~r"\s+"
+            """)
+        grammar.parse('{log (add 3 to 5)}')
+
     def test_infinite_loop(self):
         """Smoke-test a grammar that was causing infinite loops while building.
 
@@ -300,6 +325,24 @@ class GrammarTests(TestCase):
             main = number
             """)
 
+    def test_circular_toplevel_reference(self):
+        with pytest.raises(VisitationError):
+            Grammar("""
+                foo = bar
+                bar = foo
+            """)
+        with pytest.raises(VisitationError):
+            Grammar("""
+                foo = foo
+                bar = foo
+            """)
+        with pytest.raises(VisitationError):
+            Grammar("""
+                foo = bar
+                bar = baz
+                baz = foo
+            """)
+
     def test_right_recursive(self):
         """Right-recursive refs should resolve."""
         grammar = Grammar("""
@@ -461,6 +504,79 @@ class GrammarTests(TestCase):
             list(grammar.keys()),
             ['r%s' % i for i in range(100)])
 
+    def test_sequence_choice_bug(self):
+        """
+        Regression test for https://github.com/erikrose/parsimonious/issues/238
+        """
+        grammar = Grammar(r'''
+            value = "[" "]" / "5"
+        ''')
+        self.assertTrue(grammar.parse('[]') is not None)
+        self.assertTrue(grammar.parse('5') is not None)
+        grammar2 = Grammar(r'''
+            value = "5" / "[" "]"
+        ''')
+        self.assertTrue(grammar2.parse('[]') is not None)
+        self.assertTrue(grammar2.parse('5') is not None)
+        grammar3 = Grammar(r'''
+            value = "4" / "[" "]" / "(" ")" / "{" "}" / "5"
+        ''')
+        self.assertTrue(grammar3.parse('[]') is not None)
+        self.assertTrue(grammar3.parse('5') is not None)
+        self.assertTrue(grammar3.parse('()') is not None)
+        self.assertTrue(grammar3.parse('{}') is not None)
+        self.assertTrue(grammar3.parse('4') is not None)
+
+    def test_repetitions(self):
+        grammar = Grammar(r'''
+            left_missing = "a"{,5}
+            right_missing = "a"{5,}
+            exact = "a"{5}
+            range = "a"{2,5}
+            optional = "a"?
+            plus = "a"+
+            star = "a"*
+        ''')
+        should_parse = [
+            ("left_missing", ["a" * i for i in range(6)]),
+            ("right_missing", ["a" * i for i in range(5, 8)]),
+            ("exact", ["a" * 5]),
+            ("range", ["a" * i for i in range(2, 6)]),
+            ("optional", ["", "a"]),
+            ("plus", ["a", "aa"]),
+            ("star", ["", "a", "aa"]),
+        ]
+        for rule, examples in should_parse:
+            for example in examples:
+                assert grammar[rule].parse(example)
+
+        should_not_parse = [
+            ("left_missing", ["a" * 6]),
+            ("right_missing", ["a" * i for i in range(5)]),
+            ("exact", ["a" * i for i in list(range(5)) + list(range(6, 10))]),
+            ("range", ["a" * i for i in list(range(2)) + list(range(6, 10))]),
+            ("optional", ["aa"]),
+            ("plus", [""]),
+            ("star", ["b"]),
+        ]
+        for rule, examples in should_not_parse:
+            for example in examples:
+                with pytest.raises(ParseError):
+                    grammar[rule].parse(example)
+
+    def test_equal(self):
+        grammar_def = (r"""
+            x = y / z / ""
+            y = "y" x
+            z = "z" x
+        """)
+        assert Grammar(grammar_def) == Grammar(grammar_def)
+
+        self.assertEqual(Grammar(rule_syntax), Grammar(rule_syntax))
+        self.assertNotEqual(Grammar('expr = ~"[a-z]{1,3}"'), Grammar('expr = ~"[a-z]{2,3}"'))
+        self.assertNotEqual(Grammar('expr = ~"[a-z]{1,3}"'), Grammar('expr = ~"[a-z]{1,4}"'))
+        self.assertNotEqual(Grammar('expr = &"a"'), Grammar('expr = !"a"'))
+
 
 class TokenGrammarTests(TestCase):
     """Tests for the TokenGrammar class and associated machinery"""
@@ -482,14 +598,106 @@ class TokenGrammarTests(TestCase):
         grammar = TokenGrammar("""
             foo = "token1" "token2"
             """)
-        self.assertRaises(ParseError,
-                      grammar.parse,
-                      [Token('tokenBOO'), Token('token2')])
+        with pytest.raises(ParseError) as e:
+            grammar.parse([Token('tokenBOO'), Token('token2')])
+        assert "Rule 'foo' didn't match at" in str(e.value)
 
     def test_token_repr(self):
-        t = Token(u'💣')
+        t = Token('💣')
         self.assertTrue(isinstance(t.__repr__(), str))
-        if sys.version_info < (3, 0):
-            self.assertEqual(u'<Token "💣">'.encode('utf-8'), t.__repr__())
-        else:
-            self.assertEqual(u'<Token "💣">', t.__repr__())
+        self.assertEqual('<Token "💣">', t.__repr__())
+
+    def test_token_star_plus_expressions(self):
+        a = Token("a")
+        b = Token("b")
+        grammar = TokenGrammar("""
+            foo = "a"*
+            bar = "a"+
+        """)
+        assert grammar["foo"].parse([]) is not None
+        assert grammar["foo"].parse([a]) is not None
+        assert grammar["foo"].parse([a, a]) is not None
+
+        with pytest.raises(ParseError):
+            grammar["foo"].parse([a, b])
+        with pytest.raises(ParseError):
+            grammar["foo"].parse([b])
+
+        assert grammar["bar"].parse([a]) is not None
+        with pytest.raises(ParseError):
+            grammar["bar"].parse([a, b])
+        with pytest.raises(ParseError):
+            grammar["bar"].parse([b])
+
+
+def test_precedence_of_string_modifiers():
+    # r"strings", etc. should be parsed as a single literal, not r followed
+    # by a string literal.
+    g = Grammar(r"""
+        escaped_bell = r"\b"
+        r = "irrelevant"
+    """)
+    assert isinstance(g["escaped_bell"], Literal)
+    assert g["escaped_bell"].literal == "\\b"
+    with pytest.raises(ParseError):
+        g.parse("irrelevant\b")
+
+    g2 = Grammar(r"""
+        escaped_bell = r"\b"
+    """)
+    assert g2.parse("\\b")
+
+
+def test_binary_grammar():
+    g = Grammar(r"""
+        file = header body terminator
+        header = b"\xFF" length b"~"
+        length = ~rb"\d+"
+        body = ~b"[^\xFF]*"
+        terminator = b"\xFF"
+    """)
+    length = 22
+    assert g.parse(b"\xff22~" + (b"a" * 22) + b"\xff") is not None
+
+
+def test_inconsistent_string_types_in_grammar():
+    with pytest.raises(VisitationError) as e:
+        Grammar(r"""
+            foo = b"foo"
+            bar = "bar"
+        """)
+    assert e.value.original_class is BadGrammar
+    with pytest.raises(VisitationError) as e:
+        Grammar(r"""
+            foo = ~b"foo"
+            bar = "bar"
+        """)
+    assert e.value.original_class is BadGrammar
+
+    # The following should parse without errors because they use the same
+    # string types:
+    Grammar(r"""
+        foo = b"foo"
+        bar = b"bar"
+    """)
+    Grammar(r"""
+        foo = "foo"
+        bar = "bar"
+    """)
+
+
+def test_left_associative():
+    # Regression test for https://github.com/erikrose/parsimonious/issues/209
+    language_grammar = r"""
+    expression = operator_expression / non_operator_expression
+    non_operator_expression = number_expression
+
+    operator_expression = expression "+" non_operator_expression
+
+    number_expression = ~"[0-9]+"
+    """
+
+    grammar = Grammar(language_grammar)
+    with pytest.raises(LeftRecursionError) as e:
+        grammar["operator_expression"].parse("1+2")
+    assert "Parsimonious is a packrat parser, so it can't handle left recursion." in str(e.value)
